import { Component } from '@angular/core';
import { TesService } from './tes.service';
import { AuthorisationService} from './authorisation.service';
import { ComputesitesService} from './computesites.service';
import {BehaviorSubject} from 'rxjs/BehaviorSubject';
import { MatSnackBar } from '@angular/material';
import {OverlayContainer} from '@angular/cdk/overlay';
import { Computesite, Health } from './computesite';
import {SettingsService } from './settings.service';



@Component({
  selector: 'app-root',
  templateUrl: './app.component.html',
  styleUrls: ['./app.component.css']
})
export class AppComponent {
  title = 'Strudelv2';
  private loggedin: Boolean = false;
  private errmsg: string='';
  public authorised: boolean;
  public statusMsg: BehaviorSubject<string>;
  private snackBarRef: any;

  // private testingAuth: boolean;

  constructor(private tesService: TesService,
              private authService: AuthorisationService,
              private computeSitesService: ComputesitesService,
              private settingsService: SettingsService,
              public snackBar: MatSnackBar) {

   };

  ngOnInit() {
    // this.tesService.testingAuth.subscribe(testingAuth => { this.testingAuth = testingAuth; console.log('testingAuth updated'+this.testingAuth) });
    // this.testingAuth = false;
    this.statusMsg = new BehaviorSubject<any>('');
    this.tesService.setStatusMsg(this.statusMsg);
    this.computeSitesService.setStatusMsg(this.statusMsg);
    this.authService.setStatusMsg(this.statusMsg);
    this.statusMsg.subscribe(msg => this.displayMessage(msg));
    this.settingsService.theme$.subscribe((v) => this.setTheme(v))

  }

  toggleMenu() {
    if (this.settingsService.menuToggle$.value == true) {
      this.settingsService.menuToggle$.next(false);
      return
    }
    this.settingsService.menuToggle$.next(true);
  }


  private displayMessage(msg) {
   if (msg === null || msg == "") {
     if (this.snackBarRef != undefined) {
       this.snackBarRef.dismiss()
     }
   } else {
     this.snackBarRef = this.snackBar.open(msg,'Dismiss');
   }
 }
  setTheme(theme: string)  {
    if (theme == null) {
      theme = 'strudel-theme-light';
    }
    let classList = document.querySelector('app-root').classList
    var c;
    for (c of classList.value.split(' ')) {
      if (c.indexOf('strudel-') == 0) {
        console.log('remove class'+c);
        classList.remove(c);
      }
    }
    classList.add(theme);
  }

  countErrors(a: Health[], b: Health[]) {
    var count: number = 0
    var h: Health;
    if (a != null) {
      for (h of a) {
        if (h.stat == 'error' || h.stat == 'warn') {
          count++;
        }
        if(!isNaN(parseInt(h.stat))) {
          count = count + parseInt(h.stat);
        }
      }
    }
    if (b != null) {
      for (h of b) {
        if (h.stat == 'error' || h.stat == 'warn') {
          count++;
        }
        if(!isNaN(parseInt(h.stat))) {
          count = count + parseInt(h.stat);
        }
      }
    }
    return count;
  }


  // login() {
  //   console.log("login");
  //   this.authService.login()
  // }
  //
  // logout() {
  //   console.log("logout");
  //   this.tesService.logout();
  // }
}
