import { Injectable } from '@angular/core';
import { APIServer } from './apiserver';
import {BehaviorSubject} from 'rxjs';
import {take, filter} from 'rxjs/operators';

import { HttpClientModule, HttpClient, HttpHeaders, HttpErrorResponse } from '@angular/common/http';
import { environment } from '../environments/environment';

@Injectable({
  providedIn: 'root'
})
export class BackendSelectionService {

  public apiservers: BehaviorSubject<APIServer[]>;
  public apiserver: BehaviorSubject<APIServer>;

    constructor(private http: HttpClient) { 
        this.apiservers = new BehaviorSubject([]);
        this.apiserver = new BehaviorSubject(null);
        this.initApiServer();
        this.getAPIServers();
    }

    private initApiServer() {
        try {
            this.apiserver.next(<APIServer>JSON.parse(localStorage.getItem('apiserver')));
        } catch {
          this.apiserver.next(null);
        }
      if (this.apiserver.value === null) {
          //this.apiservers.pipe(filter((v) => v !== []),take(1)).subscribe((v) => {console.log('got some apiservers to choose from'); this.defaultApiServer(v)})
          this.apiservers.subscribe((v) => this.defaultApiServer(v))
      }
    }

    private defaultApiServer(v: APIServer[]) {
      this.apiserver.next(v[0]);
    }

    private saveLastApiServer(s: APIServer) {
        localStorage.setItem('apiserver', JSON.stringify(s));
    }

     setApiServer(server: APIServer) {
         this.apiserver.next(server);
         this.saveLastApiServer(this.apiserver.value)
     }

    /*     storeLocalAPIServers(apiservers) {
         localStorage.setItem('localAPIServers',JSON.stringify(apiservers));
         this.getAPIServers();
     }
     
     removeLocalAPIServer() {
         localStorage.removeItem('localAPIServers');
         this.getAPIServers();
     }*/

     getAPIServers() {
       let headers = new HttpHeaders();
       let options = { headers: headers, withCredentials: false};
       this.http.get<APIServer[]>('./assets/config/apiservers.json',options)
                        .subscribe(resp => this.updateAPIServers(resp));
     }

     private updateAPIServers(resp) {
         var s: APIServer;
         var list: APIServer[] = []
         var current: APIServer;
         var found: boolean; 
         var localServers: APIServer[] = [];
         current = this.apiserver.value;
         found = false;
         try {
             localServers = <APIServer[]>JSON.parse(localStorage.getItem('localAPIServers'));
         } catch {
             localServers = []
         }

         for (s of <APIServer[]>resp) {
             if (current !== undefined && current !== null && s.name == current.name) {
                 list.push(current);
                 found = true;
             } else {
                 list.push(s);
             }
         }
         if (localServers !== null) {
             for (s of localServers) {
                 if (current !== null && s.name == current.name) {
                     list.push(current);
                     found = true;
                 } else {
                     list.push(s);
                 }
             } 
         }
         if (!found && current !== undefined) {
             list.push(current);
         }
         this.apiservers.next(list);
     }

}
